import re
from abc import abstractmethod, ABC

from beets.dbcore import Query
from beetsplug.emd_metadata import ExtendedMetaData


class ExtendedMetaDataMatchQuery(Query):

    input_field = None

    def __init__(self, pattern):
        self.pattern = pattern

    def match(self, item):
        emd = item[ExtendedMetaDataMatchQuery.input_field]
        meta_data = ExtendedMetaData.decode(emd)

        if meta_data is None:
            return False

        query = create_query(self.pattern)

        if query is None:
            return False

        return query.matches(meta_data)


class EmdQuery(ABC):
    _unicode_letters = '\u0030-\u0039\u0041-\u005A\u0061-\u007A\u00AA\u00B5\u00BA\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02C1\u02C6-\u02D1\u02E0-\u02E4\u02EC\u02EE\u0370-\u0374\u0376\u0377\u037A-\u037D\u0386\u0388-\u038A\u038C\u038E-\u03A1\u03A3-\u03F5\u03F7-\u0481\u048A-\u0527\u0531-\u0556\u0559\u0561-\u0587\u05D0-\u05EA\u05F0-\u05F2\u0620-\u064A\u066E\u066F\u0671-\u06D3\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u06FC\u06FF\u0710\u0712-\u072F\u074D-\u07A5\u07B1\u07CA-\u07EA\u07F4\u07F5\u07FA\u0800-\u0815\u081A\u0824\u0828\u0840-\u0858\u08A0\u08A2-\u08AC\u0904-\u0939\u093D\u0950\u0958-\u0961\u0971-\u0977\u0979-\u097F\u0985-\u098C\u098F\u0990\u0993-\u09A8\u09AA-\u09B0\u09B2\u09B6-\u09B9\u09BD\u09CE\u09DC\u09DD\u09DF-\u09E1\u09F0\u09F1\u0A05-\u0A0A\u0A0F\u0A10\u0A13-\u0A28\u0A2A-\u0A30\u0A32\u0A33\u0A35\u0A36\u0A38\u0A39\u0A59-\u0A5C\u0A5E\u0A72-\u0A74\u0A85-\u0A8D\u0A8F-\u0A91\u0A93-\u0AA8\u0AAA-\u0AB0\u0AB2\u0AB3\u0AB5-\u0AB9\u0ABD\u0AD0\u0AE0\u0AE1\u0B05-\u0B0C\u0B0F\u0B10\u0B13-\u0B28\u0B2A-\u0B30\u0B32\u0B33\u0B35-\u0B39\u0B3D\u0B5C\u0B5D\u0B5F-\u0B61\u0B71\u0B83\u0B85-\u0B8A\u0B8E-\u0B90\u0B92-\u0B95\u0B99\u0B9A\u0B9C\u0B9E\u0B9F\u0BA3\u0BA4\u0BA8-\u0BAA\u0BAE-\u0BB9\u0BD0\u0C05-\u0C0C\u0C0E-\u0C10\u0C12-\u0C28\u0C2A-\u0C33\u0C35-\u0C39\u0C3D\u0C58\u0C59\u0C60\u0C61\u0C85-\u0C8C\u0C8E-\u0C90\u0C92-\u0CA8\u0CAA-\u0CB3\u0CB5-\u0CB9\u0CBD\u0CDE\u0CE0\u0CE1\u0CF1\u0CF2\u0D05-\u0D0C\u0D0E-\u0D10\u0D12-\u0D3A\u0D3D\u0D4E\u0D60\u0D61\u0D7A-\u0D7F\u0D85-\u0D96\u0D9A-\u0DB1\u0DB3-\u0DBB\u0DBD\u0DC0-\u0DC6\u0E01-\u0E30\u0E32\u0E33\u0E40-\u0E46\u0E81\u0E82\u0E84\u0E87\u0E88\u0E8A\u0E8D\u0E94-\u0E97\u0E99-\u0E9F\u0EA1-\u0EA3\u0EA5\u0EA7\u0EAA\u0EAB\u0EAD-\u0EB0\u0EB2\u0EB3\u0EBD\u0EC0-\u0EC4\u0EC6\u0EDC-\u0EDF\u0F00\u0F40-\u0F47\u0F49-\u0F6C\u0F88-\u0F8C\u1000-\u102A\u103F\u1050-\u1055\u105A-\u105D\u1061\u1065\u1066\u106E-\u1070\u1075-\u1081\u108E\u10A0-\u10C5\u10C7\u10CD\u10D0-\u10FA\u10FC-\u1248\u124A-\u124D\u1250-\u1256\u1258\u125A-\u125D\u1260-\u1288\u128A-\u128D\u1290-\u12B0\u12B2-\u12B5\u12B8-\u12BE\u12C0\u12C2-\u12C5\u12C8-\u12D6\u12D8-\u1310\u1312-\u1315\u1318-\u135A\u1380-\u138F\u13A0-\u13F4\u1401-\u166C\u166F-\u167F\u1681-\u169A\u16A0-\u16EA\u1700-\u170C\u170E-\u1711\u1720-\u1731\u1740-\u1751\u1760-\u176C\u176E-\u1770\u1780-\u17B3\u17D7\u17DC\u1820-\u1877\u1880-\u18A8\u18AA\u18B0-\u18F5\u1900-\u191C\u1950-\u196D\u1970-\u1974\u1980-\u19AB\u19C1-\u19C7\u1A00-\u1A16\u1A20-\u1A54\u1AA7\u1B05-\u1B33\u1B45-\u1B4B\u1B83-\u1BA0\u1BAE\u1BAF\u1BBA-\u1BE5\u1C00-\u1C23\u1C4D-\u1C4F\u1C5A-\u1C7D\u1CE9-\u1CEC\u1CEE-\u1CF1\u1CF5\u1CF6\u1D00-\u1DBF\u1E00-\u1F15\u1F18-\u1F1D\u1F20-\u1F45\u1F48-\u1F4D\u1F50-\u1F57\u1F59\u1F5B\u1F5D\u1F5F-\u1F7D\u1F80-\u1FB4\u1FB6-\u1FBC\u1FBE\u1FC2-\u1FC4\u1FC6-\u1FCC\u1FD0-\u1FD3\u1FD6-\u1FDB\u1FE0-\u1FEC\u1FF2-\u1FF4\u1FF6-\u1FFC\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2119-\u211D\u2124\u2126\u2128\u212A-\u212D\u212F-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2183\u2184\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2E2F\u3005\u3006\u3031-\u3035\u303B\u303C\u3041-\u3096\u309D-\u309F\u30A1-\u30FA\u30FC-\u30FF\u3105-\u312D\u3131-\u318E\u31A0-\u31BA\u31F0-\u31FF\u3400-\u4DB5\u4E00-\u9FCC\uA000-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA61F\uA62A\uA62B\uA640-\uA66E\uA67F-\uA697\uA6A0-\uA6E5\uA717-\uA71F\uA722-\uA788\uA78B-\uA78E\uA790-\uA793\uA7A0-\uA7AA\uA7F8-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8F2-\uA8F7\uA8FB\uA90A-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA60-\uAA76\uAA7A\uAA80-\uAAAF\uAAB1\uAAB5\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uABC0-\uABE2\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFA6D\uFA70-\uFAD9\uFB00-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40\uFB41\uFB43\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC'
    _word_pattern = f'[{_unicode_letters}-_ !?]+'

    @staticmethod
    @abstractmethod
    def _create(query):
        pass

    @abstractmethod
    def matches(self, metadata: ExtendedMetaData):
        pass


def create_query(raw_query):
    for s in EmdQuery.__subclasses__():
        query = s._create(raw_query)

        if query is not None:
            return query

    return None


class OneOfQuery(EmdQuery):
    CONTAINS_MATCHER = ':'
    _query_pattern = f'^({EmdQuery._word_pattern})({CONTAINS_MATCHER}((!?){EmdQuery._word_pattern}(,(!?){EmdQuery._word_pattern})*)*)?$'

    def __init__(self, field, values, negated_values):
        self.field = field
        self.values = values
        self.negated_values = negated_values

    def matches(self, metadata: ExtendedMetaData):
        if len(self.values) == 0 and len(self.negated_values) == 0:
            return self.field in metadata
        if len(self.values) == 0 and len(self.negated_values) == 1 and len(self.negated_values[0]) == 0:
            return False

        if self.field not in metadata:
            return len(self.negated_values) == 1 and len(self.values) == 0

        field_value = metadata[self.field]

        for pattern in self.values:
            for item in field_value:
                if pattern.lower() == item.lower():
                    return True

        for pattern in self.negated_values:
            for item in field_value:
                if pattern.lower() == item.lower():
                    return False

        return len(self.negated_values) > 0

    @staticmethod
    def _create(raw_query):
        re_result = re.search(OneOfQuery._query_pattern, raw_query)

        if re_result is None:
            return None

        field = re_result.group(1)
        selectors = re_result.group(3)

        if field is None:
            return None

        if selectors is None:
            return OneOfQuery(field, [], [])

        values = [v for v in selectors.split(',') if v]
        negated_values = [x[1:] for x in values if x.startswith('!')]
        values = [x for x in values if not x.startswith('!')]

        return OneOfQuery(field, values, negated_values)


class RegexQuery(EmdQuery):
    REGEX_MATCHER = '::'
    _query_pattern = f'^({EmdQuery._word_pattern}){REGEX_MATCHER}(.*)$'

    def __init__(self, field, regex):
        self.field = field
        self.regex = regex

    def matches(self, metadata: ExtendedMetaData):
        if len(self.regex) == 0:
            return True

        if self.field not in metadata:
            return False

        field_value = metadata[self.field]

        for item in field_value:
            if re.search(self.regex, item) is not None:
                return True
        return False

    @staticmethod
    def _create(raw_query):
        re_result = re.search(RegexQuery._query_pattern, raw_query)

        if re_result is None:
            return None

        field = re_result.group(1)
        regex = re_result.group(2)

        if field is None or regex is None:
            return None

        try:
            re.compile(regex)
        except re.error:
            return None

        return RegexQuery(field, regex)




import unittest

from beetsplug.extendedmetadata import ExtendedMetaDataMatchQuery
from mockito import mockito, when, mock, unstub


class ExtendedMetaDataMatchQueryTest(unittest.TestCase):

    def setUp(self):
        # {"origin":"korea","language":["korean","english"],"vocal_gender":"male", "genre":"k-pop", "tag":["two words"], "test":"two words", "jpn_word":"けいおん!"}
        self.extended_metadata = 'EMD: eyJvcmlnaW4iOiJrb3JlYSIsImxhbmd1YWdlIjpbImtvcmVhbiIsImVuZ2xpc2giXSwidm9jYWxfZ2VuZGVyIjoibWFsZSIsICJnZW5yZSI6ImstcG9wIiwgInRhZyI6WyJ0d28gd29yZHMiXSwgInRlc3QiOiJ0d28gd29yZHMiLCAianBuX3dvcmQiOiLjgZHjgYTjgYrjgpMhIn0='
        self.emd_item = self.item(self.extended_metadata)
        ExtendedMetaDataMatchQuery.input_field = 'comments'

    def tearDown(self):
        unstub()

    def query(self, val):
        return ExtendedMetaDataMatchQuery(val)

    def item(self, emd):
        item = mock()
        when(item).__getitem__('comments').thenReturn(emd)
        return item

    # INVALID PATTERNS #
    def test_invalid_pattern(self):
        self.assertFalse(self.query('language:,').match(self.emd_item))
        self.assertFalse(self.query('language:,korean').match(self.emd_item))
        self.assertFalse(self.query('language:korean,').match(self.emd_item))
        self.assertFalse(self.query('language:korean,english,').match(self.emd_item))
        self.assertFalse(self.query('language:!').match(self.emd_item))

    def test_empty_json(self):
        self.assertFalse(self.query('origin:').match(self.item('{}')))

    def test_empty_non_json(self):
        self.assertFalse(self.query('origin:').match(self.item('korea')))

    def test_empty_pattern(self):
        self.assertTrue(self.query('origin:').match(self.emd_item))

    # SINGLE VALUE #
    def test_valid_pattern(self):
        self.assertTrue(self.query('origin:korea').match(self.emd_item))
        self.assertTrue(self.query('genre:k-pop').match(self.emd_item))

    def test_different_case(self):
        self.assertTrue(self.query('origin:Korea').match(self.emd_item))

    def test_unicode_characters(self):
        self.assertTrue(self.query('jpn_word:けいおん!').match(self.emd_item))

    def test_negated(self):
        self.assertFalse(self.query('origin:!korea').match(self.emd_item))
        self.assertTrue(self.query('origin:!usa').match(self.emd_item))
        self.assertTrue(self.query('tag:!vocaloid').match(self.emd_item))

    def test_spaces(self):
        self.assertTrue(self.query('tag:two words').match(self.emd_item))
        self.assertTrue(self.query('test:two words').match(self.emd_item))

    # MULTI VALUE #
    def test_array_negated(self):
        self.assertFalse(self.query('language:!korean').match(self.emd_item))
        self.assertTrue(self.query('language:!japanese').match(self.emd_item))
        self.assertTrue(self.query('language:!korean,english').match(self.emd_item))
        self.assertFalse(self.query('language:!korean,!english,').match(self.emd_item))

    def test_array_one_of_match(self):
        self.assertTrue(self.query('language:korean,japanese').match(self.emd_item))
        self.assertTrue(self.query('language:chinese,english').match(self.emd_item))
        self.assertFalse(self.query('language:chinese,japanese').match(self.emd_item))

    def test_tag_exists(self):
        self.assertTrue(self.query('genre').match(self.emd_item))

    def test_non_existing_tag(self):
        self.assertFalse(self.query('xxx').match(self.emd_item))

    def test_array_empty_pattern(self):
        self.assertTrue(self.query('language:').match(self.emd_item))

    def test_non_existing_pattern(self):
        self.assertFalse(self.query('origin:usa').match(self.emd_item))

    def test_array_existing_pattern(self):
        self.assertTrue(self.query('language:korean').match(self.emd_item))

    def test_array_existing_pattern_different_case(self):
        self.assertTrue(self.query('language:Korean').match(self.emd_item))

    def test_array_non_existing_pattern(self):
        self.assertFalse(self.query('language:japanese').match(self.emd_item))



import base64
import json
import re
from json import JSONDecodeError


class ExtendedMetaData:
    _base64_pattern = '(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?=?'
    _metadata_pattern = f'^EMD: ({_base64_pattern})$'

    def __init__(self, data=None):
        self._data = data if data else {}

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value if isinstance(value, list) else [value]

    def __delitem__(self, key):
        del self._data[key]

    def __contains__(self, key):
        return key in self._data

    def __str__(self):
        return str(self._json_data())

    @staticmethod
    def decode(raw_value):
        re_result = re.search(ExtendedMetaData._metadata_pattern, str(raw_value))

        if re_result is None:
            return None

        encoded_metadata = re_result.group(1)

        if encoded_metadata is None:
            return None

        try:
            meta_data = json.loads(base64.b64decode(encoded_metadata))
        except JSONDecodeError:
            return None

        for tag in meta_data:
            value = meta_data[tag]
            if not isinstance(value, list):
                meta_data[tag] = [value]

        return ExtendedMetaData(meta_data)

    def encode(self):
        encoded_data = base64.b64encode(self._json_data().encode('utf-8')).decode('utf-8')
        return f'EMD: {encoded_data}'

    def _json_data(self):
        meta_data = self._data.copy()

        for tag in meta_data.copy():
            value = meta_data[tag]
            if isinstance(value, list):
                if len(value) == 1:
                    meta_data[tag] = value[0]
                elif len(value) == 0:
                    del meta_data[tag]

        return json.dumps(meta_data)



from beets.ui import Subcommand

from beetsplug.emd_metadata import ExtendedMetaData
from beetsplug.emd_command_options import EmdAddOption, EmdDeleteOption, EmdUpdateOption, EmdRenameTagOption, \
    EmdShowOption

emd_command_options = [
    EmdUpdateOption(),
    EmdRenameTagOption(),
    EmdAddOption(),
    EmdDeleteOption(),
    EmdShowOption()
]


class ExtendedMetaDataCommand(Subcommand):
    def __init__(self, input_field):
        super(ExtendedMetaDataCommand, self).__init__('emd', help=u'manage extended meta data')

        self.input_field = input_field

        self.parser.add_option(
            '-q', '--query', dest='query',
            action="store", type="string", default=[],
            help='a beets query that matches the items to which the actions will be applied to.'
        )

        for option in emd_command_options:
            option.add_parser_option(self.parser)

        self.func = self.handle_command

    def handle_command(self, lib, opts, _):
        if not self._options_are_valid(opts):
            self.parser.print_help()
            return

        query = opts.query
        items = lib.items(query)

        for item in items:
            print(item)

            old_tags = dict(item).copy()
            emd = self._get_emd(item)

            for option in emd_command_options:
                option.apply(emd, getattr(opts, option.parser_destination()))

            self._update_emd(item, emd)
            new_tags = dict(item)

            if old_tags == new_tags:
                continue

            item.write()
            item.store()

    @staticmethod
    def _options_are_valid(opts):
        if not opts.query:
            return False

        for option in emd_command_options:
            if getattr(opts, option.parser_destination()):
                return True
        return False

    def _get_emd(self, item):
        raw_emd = item[self.input_field]
        emd = ExtendedMetaData.decode(raw_emd)

        if not emd:
            emd = ExtendedMetaData()

        return emd

    def _update_emd(self, item, emd):
        updated_emd = emd.encode()
        item[self.input_field] = updated_emd



from mediafile import (MediaField, MP3StorageStyle, MP4StorageStyle, StorageStyle, ASFStorageStyle, ListMediaField, \
    MP3ListStorageStyle, MP4ListStorageStyle, ListStorageStyle, MP3PeopleStorageStyle, MP3SlashPackStorageStyle, \
MP4TupleStorageStyle, MP3DescStorageStyle, MP4BoolStorageStyle, DateField, MP3UFIDStorageStyle, MP4SoundCheckStorageStyle, MP3SoundCheckStorageStyle)

fields = {
    "title": MediaField(
        MP3StorageStyle('TIT2'),
        MP4StorageStyle('\xa9nam'),
        StorageStyle('TITLE'),
        ASFStorageStyle('Title'),
    ),
    "artist": MediaField(
        MP3StorageStyle('TPE1'),
        MP4StorageStyle('\xa9ART'),
        StorageStyle('ARTIST'),
        ASFStorageStyle('Author'),
    ),
    "album": MediaField(
        MP3StorageStyle('TALB'),
        MP4StorageStyle('\xa9alb'),
        StorageStyle('ALBUM'),
        ASFStorageStyle('WM/AlbumTitle'),
    ),
    "genres": ListMediaField(
        MP3ListStorageStyle('TCON'),
        MP4ListStorageStyle('\xa9gen'),
        ListStorageStyle('GENRE'),
        ASFStorageStyle('WM/Genre'),
    ),
    "lyricist": MediaField(
        MP3StorageStyle('TEXT'),
        MP4StorageStyle('----:com.apple.iTunes:LYRICIST'),
        StorageStyle('LYRICIST'),
        ASFStorageStyle('WM/Writer'),
    ),
    "composer": MediaField(
        MP3StorageStyle('TCOM'),
        MP4StorageStyle('\xa9wrt'),
        StorageStyle('COMPOSER'),
        ASFStorageStyle('WM/Composer'),
    ),
    "composer_sort": MediaField(
        MP3StorageStyle('TSOC'),
        MP4StorageStyle('soco'),
        StorageStyle('COMPOSERSORT'),
        ASFStorageStyle('WM/Composersortorder'),
    ),
    "arranger": MediaField(
        MP3PeopleStorageStyle('TIPL', involvement='arranger'),
        MP4StorageStyle('----:com.apple.iTunes:Arranger'),
        StorageStyle('ARRANGER'),
        ASFStorageStyle('beets/Arranger'),
    ),

    "grouping": MediaField(
        MP3StorageStyle('TIT1'),
        MP4StorageStyle('\xa9grp'),
        StorageStyle('GROUPING'),
        ASFStorageStyle('WM/ContentGroupDescription'),
    ),
    "track": MediaField(
        MP3SlashPackStorageStyle('TRCK', pack_pos=0),
        MP4TupleStorageStyle('trkn', index=0),
        StorageStyle('TRACK'),
        StorageStyle('TRACKNUMBER'),
        ASFStorageStyle('WM/TrackNumber'),
        out_type=int,
    ),
    "tracktotal": MediaField(
        MP3SlashPackStorageStyle('TRCK', pack_pos=1),
        MP4TupleStorageStyle('trkn', index=1),
        StorageStyle('TRACKTOTAL'),
        StorageStyle('TRACKC'),
        StorageStyle('TOTALTRACKS'),
        ASFStorageStyle('TotalTracks'),
        out_type=int,
    ),
    "disc": MediaField(
        MP3SlashPackStorageStyle('TPOS', pack_pos=0),
        MP4TupleStorageStyle('disk', index=0),
        StorageStyle('DISC'),
        StorageStyle('DISCNUMBER'),
        ASFStorageStyle('WM/PartOfSet'),
        out_type=int,
    ),
    "disctotal": MediaField(
        MP3SlashPackStorageStyle('TPOS', pack_pos=1),
        MP4TupleStorageStyle('disk', index=1),
        StorageStyle('DISCTOTAL'),
        StorageStyle('DISCC'),
        StorageStyle('TOTALDISCS'),
        ASFStorageStyle('TotalDiscs'),
        out_type=int,
    ),
    "lyrics": MediaField(
        MP3DescStorageStyle(key='USLT'),
        MP4StorageStyle('\xa9lyr'),
        StorageStyle('LYRICS'),
        ASFStorageStyle('WM/Lyrics'),
    ),
    "comments": MediaField(
        MP3DescStorageStyle(key='COMM'),
        MP4StorageStyle('\xa9cmt'),
        StorageStyle('DESCRIPTION'),
        StorageStyle('COMMENT'),
        ASFStorageStyle('WM/Comments'),
        ASFStorageStyle('Description')
    ),
    "bpm": MediaField(
        MP3StorageStyle('TBPM'),
        MP4StorageStyle('tmpo', as_type=int),
        StorageStyle('BPM'),
        ASFStorageStyle('WM/BeatsPerMinute'),
        out_type=int,
    ),
    "comp": MediaField(
        MP3StorageStyle('TCMP'),
        MP4BoolStorageStyle('cpil'),
        StorageStyle('COMPILATION'),
        ASFStorageStyle('WM/IsCompilation', as_type=bool),
        out_type=bool,
    ),
    "albumartist": MediaField(
        MP3StorageStyle('TPE2'),
        MP4StorageStyle('aART'),
        StorageStyle('ALBUM ARTIST'),
        StorageStyle('ALBUMARTIST'),
        ASFStorageStyle('WM/AlbumArtist'),
    ),
    "albumtype": MediaField(
        MP3DescStorageStyle(u'MusicBrainz Album Type'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz Album Type'),
        StorageStyle('MUSICBRAINZ_ALBUMTYPE'),
        ASFStorageStyle('MusicBrainz/Album Type'),
    ),
    "label": MediaField(
        MP3StorageStyle('TPUB'),
        MP4StorageStyle('----:com.apple.iTunes:Label'),
        MP4StorageStyle('----:com.apple.iTunes:publisher'),
        StorageStyle('LABEL'),
        StorageStyle('PUBLISHER'),  # Traktor
        ASFStorageStyle('WM/Publisher'),
    ),
    "artist_sort": MediaField(
        MP3StorageStyle('TSOP'),
        MP4StorageStyle('soar'),
        StorageStyle('ARTISTSORT'),
        ASFStorageStyle('WM/ArtistSortOrder'),
    ),
    "albumartist_sort": MediaField(
        MP3DescStorageStyle(u'ALBUMARTISTSORT'),
        MP4StorageStyle('soaa'),
        StorageStyle('ALBUMARTISTSORT'),
        ASFStorageStyle('WM/AlbumArtistSortOrder'),
    ),
    "asin": MediaField(
        MP3DescStorageStyle(u'ASIN'),
        MP4StorageStyle('----:com.apple.iTunes:ASIN'),
        StorageStyle('ASIN'),
        ASFStorageStyle('MusicBrainz/ASIN'),
    ),
    "catalognum": MediaField(
        MP3DescStorageStyle(u'CATALOGNUMBER'),
        MP4StorageStyle('----:com.apple.iTunes:CATALOGNUMBER'),
        StorageStyle('CATALOGNUMBER'),
        ASFStorageStyle('WM/CatalogNo'),
    ),
    "disctitle": MediaField(
        MP3StorageStyle('TSST'),
        MP4StorageStyle('----:com.apple.iTunes:DISCSUBTITLE'),
        StorageStyle('DISCSUBTITLE'),
        ASFStorageStyle('WM/SetSubTitle'),
    ),
    "encoder": MediaField(
        MP3StorageStyle('TENC'),
        MP4StorageStyle('\xa9too'),
        StorageStyle('ENCODEDBY'),
        StorageStyle('ENCODER'),
        ASFStorageStyle('WM/EncodedBy'),
    ),
    "script": MediaField(
        MP3DescStorageStyle(u'Script'),
        MP4StorageStyle('----:com.apple.iTunes:SCRIPT'),
        StorageStyle('SCRIPT'),
        ASFStorageStyle('WM/Script'),
    ),
    "language": MediaField(
        MP3StorageStyle('TLAN'),
        MP4StorageStyle('----:com.apple.iTunes:LANGUAGE'),
        StorageStyle('LANGUAGE'),
        ASFStorageStyle('WM/Language'),
    ),
    "country": MediaField(
        MP3DescStorageStyle(u'MusicBrainz Album Release Country'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz '
                        'Album Release Country'),
        StorageStyle('RELEASECOUNTRY'),
        ASFStorageStyle('MusicBrainz/Album Release Country'),
    ),
    "albumstatus": MediaField(
        MP3DescStorageStyle(u'MusicBrainz Album Status'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz Album Status'),
        StorageStyle('MUSICBRAINZ_ALBUMSTATUS'),
        ASFStorageStyle('MusicBrainz/Album Status'),
    ),
    "media": MediaField(
        MP3StorageStyle('TMED'),
        MP4StorageStyle('----:com.apple.iTunes:MEDIA'),
        StorageStyle('MEDIA'),
        ASFStorageStyle('WM/Media'),
    ),
    "albumdisambig": MediaField(
        # This tag mapping was invented for beets (not used by Picard, etc).
        MP3DescStorageStyle(u'MusicBrainz Album Comment'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz Album Comment'),
        StorageStyle('MUSICBRAINZ_ALBUMCOMMENT'),
        ASFStorageStyle('MusicBrainz/Album Comment'),
    ),
    # Release date.
    "date": DateField(
        MP3StorageStyle('TDRC'),
        MP4StorageStyle('\xa9day'),
        StorageStyle('DATE'),
        ASFStorageStyle('WM/Year'),
        year=(StorageStyle('YEAR'),)
    ),
    # *Original* release date.
    "original_date": DateField(
        MP3StorageStyle('TDOR'),
        MP4StorageStyle('----:com.apple.iTunes:ORIGINAL YEAR'),
        StorageStyle('ORIGINALDATE'),
        ASFStorageStyle('WM/OriginalReleaseYear')
    ),
    # Nonstandard metadata.
    "artist_credit": MediaField(
        MP3DescStorageStyle(u'Artist Credit'),
        MP4StorageStyle('----:com.apple.iTunes:Artist Credit'),
        StorageStyle('ARTIST_CREDIT'),
        ASFStorageStyle('beets/Artist Credit'),
    ),
    "albumartist_credit": MediaField(
        MP3DescStorageStyle(u'Album Artist Credit'),
        MP4StorageStyle('----:com.apple.iTunes:Album Artist Credit'),
        StorageStyle('ALBUMARTIST_CREDIT'),
        ASFStorageStyle('beets/Album Artist Credit'),
    ),
    # MusicBrainz IDs.
    "mb_trackid": MediaField(
        MP3UFIDStorageStyle(owner='http://musicbrainz.org'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz Track Id'),
        StorageStyle('MUSICBRAINZ_TRACKID'),
        ASFStorageStyle('MusicBrainz/Track Id'),
    ),
    "mb_releasetrackid": MediaField(
        MP3DescStorageStyle(u'MusicBrainz Release Track Id'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz Release Track Id'),
        StorageStyle('MUSICBRAINZ_RELEASETRACKID'),
        ASFStorageStyle('MusicBrainz/Release Track Id'),
    ),
    "mb_albumid": MediaField(
        MP3DescStorageStyle(u'MusicBrainz Album Id'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz Album Id'),
        StorageStyle('MUSICBRAINZ_ALBUMID'),
        ASFStorageStyle('MusicBrainz/Album Id'),
    ),
    "mb_artistid": MediaField(
        MP3DescStorageStyle(u'MusicBrainz Artist Id'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz Artist Id'),
        StorageStyle('MUSICBRAINZ_ARTISTID'),
        ASFStorageStyle('MusicBrainz/Artist Id'),
    ),
    "mb_albumartistid": MediaField(
        MP3DescStorageStyle(u'MusicBrainz Album Artist Id'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz Album Artist Id'),
        StorageStyle('MUSICBRAINZ_ALBUMARTISTID'),
        ASFStorageStyle('MusicBrainz/Album Artist Id'),
    ),
    "mb_releasegroupid": MediaField(
        MP3DescStorageStyle(u'MusicBrainz Release Group Id'),
        MP4StorageStyle('----:com.apple.iTunes:MusicBrainz Release Group Id'),
        StorageStyle('MUSICBRAINZ_RELEASEGROUPID'),
        ASFStorageStyle('MusicBrainz/Release Group Id'),
    ),
    # Acoustid fields.
    "acoustid_fingerprint": MediaField(
        MP3DescStorageStyle(u'Acoustid Fingerprint'),
        MP4StorageStyle('----:com.apple.iTunes:Acoustid Fingerprint'),
        StorageStyle('ACOUSTID_FINGERPRINT'),
        ASFStorageStyle('Acoustid/Fingerprint'),
    ),
    "acoustid_id": MediaField(
        MP3DescStorageStyle(u'Acoustid Id'),
        MP4StorageStyle('----:com.apple.iTunes:Acoustid Id'),
        StorageStyle('ACOUSTID_ID'),
        ASFStorageStyle('Acoustid/Id'),
    ),
    # ReplayGain fields.
    "rg_track_gain": MediaField(
        MP3DescStorageStyle(
            u'REPLAYGAIN_TRACK_GAIN',
            float_places=2, suffix=u' dB'
        ),
        MP3DescStorageStyle(
            u'replaygain_track_gain',
            float_places=2, suffix=u' dB'
        ),
        MP3SoundCheckStorageStyle(
            key='COMM',
            index=0, desc=u'iTunNORM',
            id3_lang='eng'
        ),
        MP4StorageStyle(
            '----:com.apple.iTunes:replaygain_track_gain',
            float_places=2, suffix=' dB'
        ),
        MP4SoundCheckStorageStyle(
            '----:com.apple.iTunes:iTunNORM',
            index=0
        ),
        StorageStyle(
            u'REPLAYGAIN_TRACK_GAIN',
            float_places=2, suffix=u' dB'
        ),
        ASFStorageStyle(
            u'replaygain_track_gain',
            float_places=2, suffix=u' dB'
        ),
        out_type=float
    ),
    "rg_album_gain": MediaField(
        MP3DescStorageStyle(
            u'REPLAYGAIN_ALBUM_GAIN',
            float_places=2, suffix=u' dB'
        ),
        MP3DescStorageStyle(
            u'replaygain_album_gain',
            float_places=2, suffix=u' dB'
        ),
        MP4StorageStyle(
            '----:com.apple.iTunes:replaygain_album_gain',
            float_places=2, suffix=' dB'
        ),
        StorageStyle(
            u'REPLAYGAIN_ALBUM_GAIN',
            float_places=2, suffix=u' dB'
        ),
        ASFStorageStyle(
            u'replaygain_album_gain',
            float_places=2, suffix=u' dB'
        ),
        out_type=float
    ),
    "rg_track_peak": MediaField(
        MP3DescStorageStyle(
            u'REPLAYGAIN_TRACK_PEAK',
            float_places=6
        ),
        MP3DescStorageStyle(
            u'replaygain_track_peak',
            float_places=6
        ),
        MP3SoundCheckStorageStyle(
            key=u'COMM',
            index=1, desc=u'iTunNORM',
            id3_lang='eng'
        ),
        MP4StorageStyle(
            '----:com.apple.iTunes:replaygain_track_peak',
            float_places=6
        ),
        MP4SoundCheckStorageStyle(
            '----:com.apple.iTunes:iTunNORM',
            index=1
        ),
        StorageStyle(u'REPLAYGAIN_TRACK_PEAK', float_places=6),
        ASFStorageStyle(u'replaygain_track_peak', float_places=6),
        out_type=float,
    ),
    "rg_album_peak": MediaField(
        MP3DescStorageStyle(
            u'REPLAYGAIN_ALBUM_PEAK',
            float_places=6
        ),
        MP3DescStorageStyle(
            u'replaygain_album_peak',
            float_places=6
        ),
        MP4StorageStyle(
            '----:com.apple.iTunes:replaygain_album_peak',
            float_places=6
        ),
        StorageStyle(u'REPLAYGAIN_ALBUM_PEAK', float_places=6),
        ASFStorageStyle(u'replaygain_album_peak', float_places=6),
        out_type=float,
    ),
    # EBU R128 fields.
    "r128_track_gain": MediaField(
        MP3DescStorageStyle(
            u'R128_TRACK_GAIN'
        ),
        MP4StorageStyle(
            '----:com.apple.iTunes:R128_TRACK_GAIN'
        ),
        StorageStyle(
            u'R128_TRACK_GAIN'
        ),
        ASFStorageStyle(
            u'R128_TRACK_GAIN'
        ),
        out_type=int,
    ),
    "r128_album_gain": MediaField(
        MP3DescStorageStyle(
            u'R128_ALBUM_GAIN'
        ),
        MP4StorageStyle(
            '----:com.apple.iTunes:R128_ALBUM_GAIN'
        ),
        StorageStyle(
            u'R128_ALBUM_GAIN'
        ),
        ASFStorageStyle(
            u'R128_ALBUM_GAIN'
        ),
        out_type=int,
    ),
    "initial_key": MediaField(
        MP3StorageStyle('TKEY'),
        MP4StorageStyle('----:com.apple.iTunes:initialkey'),
        StorageStyle('INITIALKEY'),
        ASFStorageStyle('INITIALKEY'),
    )
}



import os
import unittest

from test.beets_container import BeetsContainer


class BeetsIntegrationTest(unittest.TestCase):

    def setUp(self):
        self.beets = BeetsContainer(music_dir=self.path('beets/music'),
                                    config_dir=self.path('beets/config'),
                                    package_src_dirs=[self.path('../beetsplug')])
        self.beets.start()

    @staticmethod
    def path(rel_path):
        script_dir = os.path.dirname(__file__)
        return os.path.join(script_dir, rel_path)

    def tearDown(self):
        self.beets.stop()

    def test_missing_options_prints_help(self):
        self.assertEqual('Usage: beet emd [options]', self.beets.command('emd -a tag1:test1')[0])
        self.assertEqual('Usage: beet emd [options]', self.beets.command('emd -q artist:Title')[0])
        self.assertEqual('Usage: beet emd [options]', self.beets.command('emd --help')[0])

    def test_invalid_syntax_prints_syntax_error(self):
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -r tag1:tag2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -r tag1 tag2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -r tag1,tag2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -r tag1>tag2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -r tag1:test1/tag2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -r tag1/tag2:test2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -r tag1/tag2:'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -a tag1'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -a tag1:'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -u tag1/tag2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -u tag1:test1/tag2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -u tag1:/tag2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -u tag1:test1>tag2:test2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -u tag1:test1:tag2:test2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -u tag1:test1 tag2:test2'))
        self._prints_syntax_exception(self.beets.command('emd -q title:"Title 1" -u tag1:test1,tag2:test2'))

    @staticmethod
    def _prints_syntax_exception(command_result):
        for line in command_result:
            if 'Exception: Invalid syntax' in line:
                return True
        return False

    def test_show_emd(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1')
        self.assertTrue('"tag1": "test1"' in self.beets.command('emd -q title:"Title 1" -s')[1])
        self.assertTrue('"tag1": ["test1", "test2"]' in self.beets.command('emd -q title:"Title 1" -a tag1:test2 -s')[1])

    def test_add_new_tag(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1')

        tagged_with_test1 = self.beets.list("x:tag1:test1")
        self.assertEqual(1, len(tagged_with_test1))
        self.assertTrue(tagged_with_test1[0].endswith('Title 1'))

    def test_add_new_tag_with_spaces(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:"test 1"')

        tagged_with_test1 = self.beets.list('x:tag1:"test 1"')
        self.assertEqual(1, len(tagged_with_test1))
        self.assertTrue(tagged_with_test1[0].endswith('Title 1'))

    def test_add_value_to_existing_tag(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1')
        self.beets.command('emd -q title:"Title 1" -a tag1:test2')

        tagged_with_test1 = self.beets.list("x:tag1:test1")
        self.assertEqual(1, len(tagged_with_test1))
        self.assertTrue(tagged_with_test1[0].endswith('Title 1'))

        tagged_with_test2 = self.beets.list("x:tag1:test2")
        self.assertEqual(1, len(tagged_with_test2))
        self.assertTrue(tagged_with_test2[0].endswith('Title 1'))

    def test_add_tags_to_multiple_files(self):
        self.beets.command('emd -q title:Title -a tag1:test1')
        self.assertEqual(3, len(self.beets.list("x:tag1:test1")))

    def test_add_multiple_tag_values(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1,test2')
        self.assertEqual(1, len(self.beets.list("x:tag1:test1 x:tag1:test2")))

    def test_add_multiple_tags(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1 -a tag2:test2')
        self.assertEqual(1, len(self.beets.list("x:tag1:test1 x:tag2:test2")))

    def test_remove_tag(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1')
        self.assertEqual(1, len(self.beets.list("x:tag1")))

        self.beets.command('emd -q title:"Title 1" -d tag1')
        self.assertEqual(0, len(self.beets.list("x:tag1")))

    def test_remove_value_from_tag(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1,test2')
        self.assertEqual(1, len(self.beets.list("x:tag1:test1 x:tag1:test2")))

        self.beets.command('emd -q title:"Title 1" -d tag1:test1')
        self.assertEqual(0, len(self.beets.list("x:tag1:test1 x:tag1:test2")))
        self.assertEqual(1, len(self.beets.list("x:tag1:test2")))

    def test_remove_last_value_from_tag(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1')
        self.assertEqual(1, len(self.beets.list("x:tag1:test1")))

        self.beets.command('emd -q title:"Title 1" -d tag1:test1')
        self.assertEqual(0, len(self.beets.list("x:tag1")))

    def test_remove_tags_from_multiple_files(self):
        self.beets.command('emd -q title:Title -a tag1:test1')
        self.beets.command('emd -q title:Title -d tag1:test1')
        self.assertEqual(0, len(self.beets.list("x:tag1:test1")))

    def test_remove_multiple_tag_values(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1,test2,test3')
        self.beets.command('emd -q title:"Title 1" -d tag1:test1,test3')
        self.assertEqual(1, len(self.beets.list("x:tag1:test2 x:tag1:!test1 x:tag1:!test3")))

    def test_remove_multiple_tag(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1 -a tag2:test2 -a tag3:test3')
        self.beets.command('emd -q title:"Title 1" -d tag1 -d tag3')
        self.assertEqual(1, len(self.beets.list("x:tag2")))
        self.assertEqual(0, len(self.beets.list("x:tag1")))
        self.assertEqual(0, len(self.beets.list("x:tag3")))

    def test_rename_tag(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1')
        self.beets.command('emd -q title:"Title 1" -r tag1/tag2')
        self.assertEqual(1, len(self.beets.list("x:tag2")))
        self.assertEqual(0, len(self.beets.list("x:tag1")))

    def test_update_tag_value(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1')
        self.beets.command('emd -q title:"Title 1" -u tag1:test1/tag1:test2')
        self.assertEqual(1, len(self.beets.list("x:tag1:test2")))
        self.assertEqual(0, len(self.beets.list("x:tag1:test1")))

    def test_move_tag_value(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1')
        self.beets.command('emd -q title:"Title 1" -u tag1:test1/tag2:test1')
        self.assertEqual(1, len(self.beets.list("x:tag2:test1")))
        self.assertEqual(0, len(self.beets.list("x:tag1:test1")))

    def test_move_and_update_tag_value(self):
        self.beets.command('emd -q title:"Title 1" -a tag1:test1')
        self.beets.command('emd -q title:"Title 1" -u tag1:test1/tag2:test2')
        self.assertEqual(1, len(self.beets.list("x:tag2:test2")))
        self.assertEqual(0, len(self.beets.list("x:tag1:test1")))



from beets.plugins import BeetsPlugin
from beetsplug.emd_command import ExtendedMetaDataCommand
from beetsplug.emd_query import ExtendedMetaDataMatchQuery


class ExtendedMetaDataPlugin(BeetsPlugin):

    def __init__(self):
        super(ExtendedMetaDataPlugin, self).__init__()

        self.input_field = self.config['input_field'].get('comments')
        self.query_prefix = self.config['query_prefix'].get('x')

        ExtendedMetaDataMatchQuery.input_field = self.input_field
        self.item_queries = {self.query_prefix: ExtendedMetaDataMatchQuery}

    def commands(self):
        return [ExtendedMetaDataCommand(self.input_field)]



import re
from abc import ABC, abstractmethod

from beetsplug.emd_query import OneOfQuery, EmdQuery


class EmdCommandOption(ABC):

    @abstractmethod
    def add_parser_option(self, parser):
        pass

    @abstractmethod
    def parser_destination(self):
        pass

    @abstractmethod
    def apply(self, emd, opt_values):
        pass


class EmdShowOption(EmdCommandOption):

    def apply(self, emd, opt_values):
        if opt_values:
            print(emd)

    def add_parser_option(self, parser):
        parser.add_option(
            '-s', '--show', dest=self.parser_destination(),
            action="store_true", help='show the extended meta data of the files'
        )

    def parser_destination(self):
        return 'show_emd'


class EmdAddOption(EmdCommandOption):

    def apply(self, emd, opt_values):
        added_tag_values_tuples = [self._extract_tag_value_tuple(v) for v in opt_values]
        self._add_tag_values(emd, added_tag_values_tuples)

    def add_parser_option(self, parser):
        parser.add_option(
            '-a', '--add', dest=self.parser_destination(),
            action="append", type="string", default=[],
            help='add a tag value. Example: "tag1:v1" or "tag1:v1,v2,v3".'
        )

    def parser_destination(self):
        return 'add_expression'

    @staticmethod
    def _extract_tag_value_tuple(value):
        query = OneOfQuery._create(value)

        if not query:
            raise Exception(f"Invalid syntax '{value}'.")

        return query.field, query.values

    @staticmethod
    def _add_tag_values(emd, value_tuples):
        if not value_tuples:
            return

        for tag, values in value_tuples:
            if tag in emd:
                tag_values = emd[tag]

                for value in values:
                    if value not in tag_values:
                        tag_values.append(value)
            else:
                emd[tag] = values


class EmdDeleteOption(EmdCommandOption):

    def apply(self, emd, opt_values):
        removed_tag_values_tuples = [self._extract_tag_value_tuple(v) for v in opt_values]
        self._remove_tag_values(emd, removed_tag_values_tuples)

    def add_parser_option(self, parser):
        parser.add_option(
            '-d', '--delete', dest=self.parser_destination(),
            action="append", type="string", default=[],
            help='delete a tag value or tag. Example: "tag1" or "tag1:v1".'
        )

    def parser_destination(self):
        return 'delete_expression'

    @staticmethod
    def _extract_tag_value_tuple(value):
        query = OneOfQuery._create(value)

        if not query:
            raise Exception(f"Invalid syntax '{value}'.")

        return query.field, query.values

    @staticmethod
    def _remove_tag_values(emd, value_tuples):
        if not value_tuples:
            return

        for tag, values in value_tuples:
            if tag not in emd:
                return

            if len(values) == 0:
                del emd[tag]
            else:
                for value in values:
                    emd[tag].remove(value)


class EmdRenameTagOption(EmdCommandOption):

    def apply(self, emd, opt_values):
        renamed_tag_tuples = [self._extract_source_destination_tag_tuple(v) for v in opt_values]
        self._rename_tags(emd, renamed_tag_tuples)

    def add_parser_option(self, parser):
        parser.add_option(
            '-r', '--rename', dest=self.parser_destination(),
            action="append", type="string", default=[],
            help='rename a tag. Example: "tag1/tag2".'
        )

    def parser_destination(self):
        return 'rename_expression'

    @staticmethod
    def _rename_tags(emd, tags):
        for old_tag, new_tag in tags:
            if old_tag in emd:
                emd[new_tag] = emd[old_tag]
                del emd[old_tag]

    @staticmethod
    def _extract_source_destination_tag_tuple(value):
        re_result = re.search(f'^({EmdQuery._word_pattern})+/({EmdQuery._word_pattern})+$', value)

        if not re_result:
            raise Exception(f"Invalid syntax '{value}'")

        return re_result.group(1), re_result.group(2)


class EmdUpdateOption(EmdCommandOption):

    def apply(self, emd, opt_values):
        updated_tag_value_tuples = [self._extract_source_destination_tag_value_tuple(v) for v in opt_values]
        self._update_tag_values(emd, updated_tag_value_tuples)

    def add_parser_option(self, parser):
        parser.add_option(
            '-u', '--update', dest=self.parser_destination(),
            action="append", type="string", default=[],
            help='update or move a tag value. Example: "tag1:v1/tag1:v2" or "tag1:v1/tag2:v1" or "tag1:v1/tag2:v2".'
        )

    def parser_destination(self):
        return 'update_expression'

    @staticmethod
    def _update_tag_values(emd, tag_value_tuples):
        for src, dst in tag_value_tuples:
            old_tag = src[0]
            new_tag = dst[0]
            old_value = src[1][0]
            new_value = dst[1][0]

            if old_tag not in emd or old_value not in emd[old_tag]:
                continue

            emd[old_tag].remove(old_value)

            if new_tag in emd:
                emd[new_tag].append(new_value)
            else:
                emd[new_tag] = new_value

    @staticmethod
    def _extract_source_destination_tag_value_tuple(value):
        re_result = re.search(f'^(.+)+/(.+)+$', value)

        if not re_result:
            raise Exception(f"Invalid syntax '{value}'")

        return EmdUpdateOption._extract_tag_value_tuple(
            re_result.group(1)), EmdUpdateOption._extract_tag_value_tuple(re_result.group(2))

    @staticmethod
    def _extract_tag_value_tuple(value):
        query = OneOfQuery._create(value)

        if not query:
            raise Exception(f"Invalid syntax '{value}'.")

        return query.field, query.values


from setuptools import setup

setup(
    name='beets-extended-metadata',
    version='0.2.4',
    description='beets plugin to use custom, extended metadata',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Joscha Düringer',
    author_email='joscha.dueringer@beardbot.net',
    url='https://github.com/calne-ca/beets-plugin-extended-metadata',
    license='MIT',
    platforms='ALL',

    test_suite='test',

    packages=['beetsplug'],

    install_requires=[
        'beets>=1.4.9',
        'mediafile>=0.9.0'
    ],

    tests_require=[
        'pytest==6.2.5',
        'mockito==1.3.0',
        'testcontainers==3.4.2'
    ],

    classifiers=[
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Sound/Audio :: Players :: MP3',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



import unittest

from beetsplug.extendedmetadata import ExtendedMetaDataMatchQuery
from mockito import mock, unstub, when


class ExtendedMetaDataMatchQueryTest(unittest.TestCase):

    def setUp(self):
        # {"origin":"korea","language":["korean","english"],"vocal_gender":"male", "genre":"k-pop", "tag":["two words"], "test":"two words", "jpn_word":"けいおん!"}
        self.extended_metadata = 'EMD: eyJvcmlnaW4iOiJrb3JlYSIsImxhbmd1YWdlIjpbImtvcmVhbiIsImVuZ2xpc2giXSwidm9jYWxfZ2VuZGVyIjoibWFsZSIsICJnZW5yZSI6ImstcG9wIiwgInRhZyI6WyJ0d28gd29yZHMiXSwgInRlc3QiOiJ0d28gd29yZHMiLCAianBuX3dvcmQiOiLjgZHjgYTjgYrjgpMhIn0='
        self.emd_item = self.item(self.extended_metadata)
        ExtendedMetaDataMatchQuery.input_field = 'comments'

    def tearDown(self):
        unstub()

    def query(self, val):
        return ExtendedMetaDataMatchQuery(val)

    def item(self, emd):
        item = mock()
        when(item).__getitem__('comments').thenReturn(emd)
        return item
    
    # SINGLE VALUE #
    def test_invalid_pattern(self):
        self.assertFalse(self.query('language::^$^').match(self.emd_item))

    def test_empty_pattern(self):
        self.assertTrue(self.query('origin::').match(self.emd_item))

    def test_existing_pattern(self):
        self.assertTrue(self.query('origin::^.*ea$').match(self.emd_item))

    def test_non_existing_pattern(self):
        self.assertFalse(self.query('origin::us.*').match(self.emd_item))

    # MULTI VALUE #
    def test_array_empty_pattern(self):
        self.assertTrue(self.query('language::').match(self.emd_item))

    def test_array_existing_pattern(self):
        self.assertTrue(self.query('language::^.*an$').match(self.emd_item))

    def test_array_non_existing_pattern(self):
        self.assertFalse(self.query('language::ja.*').match(self.emd_item))
